/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib.inferenz;

import java.util.List;

import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.Term;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Variable;

/**
 * Ein fuzzy-logisches Literal (ein Term mit Wert).
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class TermLiteral implements Variable {

	/** Der Name der Fuzzy-Variable im Kontext. */
	private final String fuzzyVariable;

	/** Name des Terms der Fuzzy-Variable des das Symbol darstellt. */
	private final String term;

	/** Wert des Literals. */
	private final Float zugehoerigkeit;

	/**
	 * Konstruiert ein Terminalsymbol mit dem angegebenen Namen.
	 *
	 * @param fuzzyVariable
	 *            Name einer Fuzzy-Variable
	 * @param term
	 *            Name eines Terms der Fuzzy-Variable
	 * @param zugehoerigkeit
	 *            Der Zugeh&ouml;rigkeitswert des Literals
	 * @throws de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterException
	 *             Wenn der Name der Fuzzy-Variable oder des Terms {@code null} ist
	 *             oder nur aus Leerzeichen besteht
	 */
	public TermLiteral(final String fuzzyVariable, final String term, final Float zugehoerigkeit) {
		Kontext.pruefeName(fuzzyVariable);
		this.fuzzyVariable = fuzzyVariable.trim();
		this.term = term;
		this.zugehoerigkeit = zugehoerigkeit;
	}

	/**
	 * Konstruiert ein Terminalsymbol mit dem angegebenen Namen.
	 *
	 * @param term
	 *            Ein Term
	 */
	public TermLiteral(final Term term) {
		fuzzyVariable = term.getFuzzyVariable().getName();
		this.term = term.getName();
		zugehoerigkeit = term.getZugehoerigkeit();
	}

	/**
	 * Gibt immer {@code null} zur&uuml;ck, da dies ein Terminalsymbol ist.
	 * <p>
	 * {@inheritDoc}
	 */
	@Override
	public List<Ausdruck> getNachfolger() {
		return null;
	}

	/**
	 * Nennt den Namen der Fuzzy-Variable, zu der dieser Term geh&ouml;rt.
	 *
	 * @return Variablenname
	 */
	@Override
	public String getName() {
		assert (fuzzyVariable != null) && !fuzzyVariable.isEmpty();
		return fuzzyVariable;
	}

	/**
	 * Nennt den Namen des Terms.
	 *
	 * @return Variablenname
	 */
	public String getTerm() {
		assert (term != null) && !term.isEmpty();
		return term;
	}

	/**
	 * Gibt die Zugeh&ouml;rigkeit des Termliterals zur&uuml;ck.
	 *
	 * @return Wert der Zugeh&ouml;rigkeit
	 */
	public Float getZugehoerigkeit() {
		return zugehoerigkeit;
	}

	/**
	 * Gibt den Term zur&uuml;ck, den dieses Symbol darstellt.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public Object interpret(final Kontext kontext) {
		FuzzyVariable fv;

		fv = (FuzzyVariable) kontext.get(getName(), FuzzyVariable.class);

		return fv.getTerm(getTerm());
	}

	/**
	 * Gibt den Namen der Fuzzy-Variable und des Terms zur&uuml;ck.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return getName() + "{" + getTerm() + "}";  
	}
}
